//
//			Yasuharu Okamoto, 2015.12
//
#ifndef CHISQRDISTRI_CK
#define CHISQRDISTRI_OK

#include "adapIntgrtn.h"

using namespace adapInt;

namespace ChiSqrD {

	inline double sqr(double v){
		return v * v;
	}

	class Chi0 {
	public:
		double df;
		Chi0(){}
		Chi0(double vdf){
			df = vdf;
		}
		double operator()(double chi){
			double v = ((df / 2.0) - 1.0) * Math::Log(chi) - (chi / 2.0);
			return Math::Exp(v);
		}
	};

	class Chi0t{
	public:
		double df;
		Chi0 chi0;
		Chi0t(){}
		Chi0t(double vdf){
			df = vdf;
			chi0.df = df;
		}
		double operator()(double t){
			double v = chi0(Math::Tan(t)) / sqr(Math::Cos(t));
			return v;
		}
	};

	class CumChi2 {
	public:
		double df;
		Chi0 chi0;
		double coeff;
		CumChi2(){}
		CumChi2(double vdf){
			df = vdf;
			chi0.df = df;
			Chi0t chi0t(df);
			TAdaptiveGL<Chi0t> gl(chi0t);
			double IntgrlChi0 = gl.integral(0.0, Math::PI / 2.0);
			coeff = 1.0 / IntgrlChi0;
		}
		void setCumChi2(double vdf){
			df = vdf;
			chi0.df = df;
			Chi0t chi0t(df);
			TAdaptiveGL<Chi0t> gl(chi0t);
			double IntgrlChi0 = gl.integral(0.0, Math::PI / 2.0);
			coeff = 1.0 / IntgrlChi0;
		}
		double operator()(double chi2){
			if (df < 1.0) throw gcnew Exception("df = " + df.ToString() + " < 1.0");
			TAdaptiveGL<Chi0> gl(chi0);
			double v = gl.integral(0.0, chi2);
			return v * coeff;
		}
	};

	//     Bisection@ɂ鍪̌vZ    
	template<class TF>
	void TBisection(TF & f,
		double s,					 //   f֐̂Ƃ s > 0.0
		double c,					 //   ɑ΂֐l
		double L_b, double U_b,		 //   ̑݋Ԃ̉Ə
		double & Root,				 //     f(Root) = c
		double acc){
		if ((f(L_b) - c) * (f(U_b) - c) > 0.0){
			throw  gcnew Exception("Ɖ̑gݍ킹sK؂łB");
		}
		do {
			double m = 0.5 * (L_b + U_b);
			double v = f(m);
			if (s * (v - c) > 0.0) {
				U_b = m;
			}
			else {
				L_b = m;
			}
			System::Diagnostics::Debug::WriteLine("L_b = " + L_b.ToString() + "   m = " + m.ToString() + "   U_b = " + U_b.ToString());
		} while ((Math::Abs(U_b - L_b) >= acc * Math::Abs(U_b))
			&&
			((Math::Abs(U_b) + Math::Abs(L_b)) >= 1.0e-14)
			);
		Root = 0.5 * (L_b + U_b);
	}

	//     Newton@ɂ鍪̌vZ   
	template<class TF, class TDF>
	void TNewton(double & x,		//  ̏lA߂l͍ f(x)=c
		double c,					//  ֐l
		TF & f,						//  ֐
		TDF & df,					//  ֐
		double acc){
		int NStep = 0;
		int ck = 0;
		double x0;
		do {
			NStep++;
			if (NStep > 20) {			//  20ȏJԂ̂Ƃ
				ck = 1;
				break;
			};
			x0 = x;
			x = x0 + (c - f(x0)) / df(x0);
			if (x <= 0.0) x = 0.5 * x0;
			if (x > x0 * 2.0) x = x0 * 2.0;
		} while ((Math::Abs(x - x0) >= (acc * (Math::Abs(x) + Math::Abs(x0))))
			&&
			((Math::Abs(x) + Math::Abs(x0)) >= 1.0e-14)
			);

		if ((ck == 0) && (x > 1.0e-16)) return;

		//   Bisection@ŌvZ

		double L_b = 1.0;
		do {
			L_b *= 0.5;
		} while (f(L_b) >= c);

		double U_b = 1.0;
		do {
			U_b += 1.0;
		} while (f(U_b) <= c);

		TBisection<TF>(f, 1.0, c, L_b, U_b, x, acc);
		return;
	}

	class Chi2_density{
	public:
		double df;
		double coeff;
		Chi0 chi0;
		Chi2_density(){}
		Chi2_density(double vdf){
			df = vdf;
			chi0.df = df;
			Chi0t chi0t(df);
			TAdaptiveGL<Chi0t> gl(chi0t);
			double IntgrlChi0 = gl.integral(0.0, Math::PI / 2.0);
			coeff = 1.0 / IntgrlChi0;
		}
		void set_df(double vdf){
			df = vdf;
			chi0.df = df;
			Chi0t chi0t(df);
			TAdaptiveGL<Chi0t> gl(chi0t);
			double IntgrlChi0 = gl.integral(0.0, Math::PI / 2.0);
			coeff = 1.0 / IntgrlChi0;
		}
		double operator()(double chi2){
			double v = chi0(chi2);
			return v * coeff;
		}
	};
	

	class Calc_chi2_for_p {
	public:
		double df;
		CumChi2 cum_chi2;
		Chi2_density chi2_density;
		Calc_chi2_for_p(){}
		Calc_chi2_for_p(double vdf){
			df = vdf;
			cum_chi2.setCumChi2(df);
			chi2_density.set_df(df);
		}
		double operator()(double p){
			double x2 = 1.0;
			TNewton<CumChi2, Chi2_density>(x2, p, cum_chi2, chi2_density, 1.0e-12);

			return x2;
		}
	};

}

#endif
